.graph {
  height:300px; 
  width: 100%;
}

#modal_background {
  background-color: var(--md-default-fg-color--lightest); 
  overflow: hidden; 
  display: flex; 
  position: fixed; 
  top: 0; 
  left: 0; 
  width: 100%; 
  height: 100%; 
  z-index: 99;
}

.modal_graph {
  border: 1px solid var(--md-default-fg-color--lightest); 
  box-shadow: 0px 0px 10px var(--md-default-fg-color--lightest); 
  border-radius: 8px; 
  overflow: hidden; 
  display: flex; 
  align-items: center; 
  justify-content: center; 
  position: absolute; 
  top: 10%; 
  left: 15%; 
  width: 70%; 
  height: 80%; 
  z-index: 100;
}